#!/usr/bin/perl

use strict;
use warnings;
use FindBin qw($Bin);
use lib "$Bin/modules";
use lib "$Bin/";

# The following block of code is used to test whether or not the
# GD::Graph module is available. This module is used to generate a
# graphical representation of the FSC curve and is not essential to
# the proper function of auto3dem. Note that the module is loaded
# using "require" rather than "use" so that we can do a run time check
# on the availability of the module.

my $gdflag;
eval {
    require GD::Graph::lines;
};
if($@) {
    $gdflag = 0;
} else {
    $gdflag = 1;
}

use File::Copy;
use Cwd;
use make_program_input;
use init_params;
use write_params;
use run_mpi_prog;
use incr;
use validate;
use psf_parse;
use write_header;
use update_res;
use get_info;
use select;
use get_ann_lo_hi;
use sendmail;
use sanity_checks;
use mode;

my $time_start = time();
my $startdir = `pwd`;
my $mode = mode();
my $version_string = "------ AUTO3DEM version 3.07 ($mode) ------";
chomp($startdir);

# Print usage info if no arguments provided
if(@ARGV == 0 || $ARGV[0] =~ /^-help/i) {
print << "EOF";
$version_string

NAME
  auto3dem - automated 3D image reconstruction program

EXAMPLE
  auto3dem -ncpu 12 -input input_file -nodefile mynodelist

DESCRIPTION

  All input specified using the syntax -key value. -ncpu (MPI version
  only) and -input required; -nodefile optional. 'n' = integer value,
  's' = string

  -input s     file containing auto3dem input parameters
  -ncpu n      number of CPUs
  -nodefile s  file containing list of compute nodes

NOTES
  Nodefile specification depends on your computing environment and in
  many cases it can be omitted. For systems running the PBS scheduler,
  the nodefile is automatically obtained from the environment assuming
  that "#PBS -V" was used in the batch script. Keys are case
  insensitive (e.g. -ncpu, -Ncpu, -NCPU all OK), but string values are
  case sensitive.
EOF
exit(0);
}

# Process command line args so that we can split on ' -'
my $command_line = join ' ', @ARGV;
$command_line =~ s/^\s+//;   # Get rid of leading whitespace
$command_line =~ s/\s+$//;   # Get rid of trailing whitespace
$command_line =~ s/=/ /g;    # Change '=' to ' '
$command_line =~ s/^-//;     # Get rid of leading '-'
@ARGV = split(/\s+-/, $command_line);

my %cmdline_arg;
foreach my $arg (@ARGV) {
    my($key, $value) = split(/\s+/, $arg);
    $key =~ tr/A-Z/a-z/;
    $cmdline_arg{$key} = $value;
}

my ($ncpu, $nodefile, $param_file);

if(defined $cmdline_arg{ncpu}) {
    $ncpu = $cmdline_arg{ncpu};
}

if(defined $cmdline_arg{nodefile}) {
    $nodefile = $cmdline_arg{nodefile};
    unless($nodefile =~ /^\//) {
	$nodefile = "$startdir/$nodefile";
    }
} elsif (defined $ENV{PBS_NODEFILE}) {
    $nodefile = $ENV{PBS_NODEFILE};
} elsif (defined $ENV{SGE_NODEFILE}) {
    $nodefile = $ENV{SGE_NODEFILE};
}

if(defined $cmdline_arg{input}) {
    $param_file = $cmdline_arg{input};
}

if($mode eq 'serial') {
    $ncpu = 1;
}
validate_commandline_args($ncpu, $param_file, $mode);

# Initialize PO2R, P3DR, PCUT, PSF, PPFT, PCTFR, and control hashes
my (%auto_param, %po2r_param, %p3dr_param,
    %pcut_param, %psf_param, %ppft_param, %pctfr_param);
my @ptlefiles = ();
init_params($param_file, \%auto_param, \%po2r_param,
	    \%p3dr_param, \%pcut_param, \%psf_param,
	    \%ppft_param, \%pctfr_param, \@ptlefiles);

# Open outfile and write header information
if($auto_param{restart}) {
    open(OUT, "+>>$auto_param{outfile}_log");
} else {
    open(OUT, ">$auto_param{outfile}_log");
}

print OUT "$version_string\n\n";
write_header($ncpu, $nodefile, \%auto_param, 
	     \%p3dr_param, \%po2r_param, 
	     \%pcut_param, \%psf_param, \%ppft_param);

# Open summary output file and fill array of estimated resolutions if
# this is a restart calculation

my @estres = ();

my $summary_file = "$auto_param{outfile}_summary";
if($auto_param{restart} && -e $summary_file) {
    open(SUM, "+>>$summary_file");
    seek(SUM, 0, 0); # Move to start of file
    while (<SUM>) {
	s/^\s+//;
	my ($iter, $mode, $estres) = split(/\s+/);
	if(defined $iter && defined $mode && defined $estres &&
	   $iter =~ /\d/ && ($mode =~ /search/ || $mode =~ /refine/)) {
	    $estres =~ s/\*//; # Get rid of asterisk
	    unshift(@estres, $estres);
	}
    }
} else {
    open(SUM, ">$summary_file");
    print SUM "$version_string\n\n";
    print SUM "iter:         iteration number\n";
    print SUM "              0 = starting map constructed from images\n";
    print SUM "mode:         orientation refinement mode\n";
    print SUM "              number following search = bin factor\n";
    print SUM "est res:      estimated resolution of model\n";
    print SUM "              * = FSC never < $auto_param{fsc_hithresh}\n";
    print SUM "delta angle:  spacing between projections (degrees)\n";
    print SUM "map undamped: resolution to which map is computed\n";
    print SUM "map damped:   resolution at which map density damped to zero\n";
    print SUM "time:         wallclock time for iteration\n";
    print SUM "nptles:       number ptles used to construct map\n";
    print SUM "ncpu:         number of CPUs (MPI processes)\n";
    print SUM "\n";
    print SUM "All resolutions expressed in Angstroms, times in seconds\n";
    print SUM "\n";

    print SUM "                       delta    map     map      ";
    print SUM "                  \n";
    print SUM "iter mode      est res angle  undamped damped    ";
    print SUM "time   nptles ncpu\n";
    print SUM "---- --------- ------- ------ -------- --------- ";
    print SUM "-----  ------ ----\n";
}

# Change to the run directory where particle (and optionally map data)
# reside and where we want the results files (new particle
# orientations and maps) to be written

chdir($auto_param{rundir});

# Make sure that all of the particle parameter files are readable and
# properly formed

my $exit_flag = 0;

foreach my $file (@ptlefiles) {
    unless(-r $file) {
	print "\n--- Errors for file $file --- \n";
	print "Particle parameter file is missing or not readable\n";
	$exit_flag = 1;
	next;
    }
    my ($warnings_ref, $errors_ref) = 
	param_file_check($file, $auto_param{noctf});
    if(@{$warnings_ref}) {
	print "\n--- Warnings for file $file ---\n";
	foreach my $warning (@{$warnings_ref}) {
	    print "$warning\n";
	}
    }
    if(@{$errors_ref}) {
	print "\n--- Errors for file $file --- \n";
	foreach my $error (@{$errors_ref}) {
	    print "$error\n";
	}
	$exit_flag = 1;
    }
}

# Exit if any errors were encountered
if($exit_flag) {
    exit(0);
}

# Create a directory where intermediate files can be moved
my $intfiles_dir = "$auto_param{outfile}_INTFILES";
mkdir $intfiles_dir;

# Get pixel_size from first ptlefile and update PSF pixel_size
my $pixel_size = get_pixel_size($ptlefiles[0]);
$psf_param{pixel_size} = $pixel_size;

# Extract boxrad (packRadius in PIF standard) from image file
# Make undefined if value equal to zero
my $pif_name = get_pif_name($ptlefiles[0]);
my $boxrad_from_pif = get_boxrad($pif_name);
if($boxrad_from_pif == 0) {
    undef $boxrad_from_pif;
}

# The following code is used to allow for the specification of the box
# radius in either the PIF file or the auto3dem parameter
# file. Necessary because some PIF files are missing the box radius

# Case 1 - box radius not set in either location
unless(defined $auto_param{boxrad} || defined $boxrad_from_pif) {
    print OUT "Box radius not specified in PIF file or parameter file\n";
    print OUT "Either fix $pif_name or manually specify box radius\n";
    die "Terminating execution\n";
}

# Case 2 - box radius specified in both locations
if(defined $auto_param{boxrad} && defined $boxrad_from_pif) {
    print OUT "Box radius specified in both PIF file and parameter file\n";
    if($auto_param{boxrad} == $boxrad_from_pif) {
	print OUT "Box radius ($boxrad_from_pif) specified consistently\n\n";
    } else {
	print OUT "\nWarning boxrad from PIF file ($boxrad_from_pif) ";
	print OUT "not equal to boxrad from parameter file ";
	print OUT "($auto_param{boxrad})\n";
	print OUT "Value in parameter file overides PIF file\n\n";
    }
}

# Case 3 - box radius specified only in PIF file
if(defined $boxrad_from_pif && !defined $auto_param{boxrad}) {
    print OUT "Box radius from PIF file = $boxrad_from_pif\n\n";
    $auto_param{boxrad} = $boxrad_from_pif;
}

# Case 4 - box radius specified only in parameter file
if(!defined $boxrad_from_pif && defined $auto_param{boxrad}) {
    print OUT "Box radius from parameter file = $auto_param{boxrad}\n\n";
}

print OUT "Pixel size from $ptlefiles[0] = $pixel_size\n\n";
print OUT "\n";

# Use boxrad to set program parameters

unless(defined $pcut_param{in_rad}) {
    $pcut_param{in_rad} = sprintf "%0.0f", $auto_param{boxrad}/3.0;
}

unless(defined $pcut_param{out_rad}) {
    $pcut_param{out_rad} = sprintf "%0.0f", $auto_param{boxrad} - 10.0;
}

unless(defined $ppft_param{annulus_low}) {
    $ppft_param{annulus_low} = sprintf "%0.0f", $auto_param{boxrad}/3.0;
}

unless(defined $ppft_param{annulus_high}) {
    $ppft_param{annulus_high} = sprintf "%0.0f", $auto_param{boxrad} - 10.0;
}

unless(defined $ppft_param{pftrad_hi}) {
    $ppft_param{pftrad_hi} = $auto_param{boxrad};
}

unless(defined $ppft_param{resolution_low}) {
    $ppft_param{resolution_low} = $auto_param{boxrad}*2.0*$pixel_size/5.0;
}

unless(defined $po2r_param{res_min}) {
    $po2r_param{res_min} = $auto_param{boxrad}*2.0*$pixel_size/5.0;
}

unless(defined $pctfr_param{res_min}) {
    $pctfr_param{res_min} = $auto_param{boxrad}*2.0*$pixel_size/5.0;
}

# Set the number of CPUs to be used by each program
# ($PROG_param{max_cpu} is the upper limit)
my ($p3dr_cpu, $po2r_cpu, $pctfr_cpu, $pcut_cpu, $psf_cpu, $ppft_cpu);
$p3dr_cpu  = ($p3dr_param{max_cpu}  < $ncpu) ? $p3dr_param{max_cpu}  : $ncpu;
$po2r_cpu  = ($po2r_param{max_cpu}  < $ncpu) ? $po2r_param{max_cpu}  : $ncpu;
$pctfr_cpu = ($pctfr_param{max_cpu} < $ncpu) ? $pctfr_param{max_cpu} : $ncpu;
$pcut_cpu  = ($pcut_param{max_cpu}  < $ncpu) ? $pcut_param{max_cpu}  : $ncpu;
$psf_cpu   = ($psf_param{max_cpu}   < $ncpu) ? $psf_param{max_cpu}   : $ncpu;
$ppft_cpu  = ($ppft_param{max_cpu}  < $ncpu) ? $ppft_param{max_cpu}  : $ncpu;

# Determine the starting and ending iteration numbers
my ($first_iter, $last_iter);
if($auto_param{have_map}) {
    $last_iter = $auto_param{iter_start} + $auto_param{niter} - 1;
} else {
    $auto_param{iter_start}--;
    $last_iter = $auto_param{iter_start} + $auto_param{niter};
}
$first_iter = $auto_param{iter_start};

# Save the originally specified number of iterations
my $niter_original = $auto_param{niter};

# Set things up for hollowing map
my %hollow_param;
if($auto_param{hollow_map}) {
    $hollow_param{out_file_name} = "hollow_map";
    $hollow_param{in_rad}        = $auto_param{hollow_in_rad};
    $hollow_param{out_rad}       = $auto_param{hollow_out_rad};
    $hollow_param{cut_step}      = $auto_param{hollow_cut_step}; 
    $hollow_param{cut_weight}    = $auto_param{hollow_cut_weight}; 
}

# Set the restart flag to true
$auto_param{restart} = 1;

# Maintain list of map files
my @mapfiles = ();
my $map_iter;

# LOOP OVER ITERATIONS
for (my $iter=$first_iter; $iter<=$last_iter; $iter++) { 
    my $time_iter_start = time();
    my $nptles = 0;
    $map_iter = "$auto_param{outfile}_iter_${iter}.pif";

    print OUT "*********** Iteration $iter ***********\n\n";

    my $input_file;
    my $output_file;
    
    $auto_param{iter_start}++;

    # Run PO2R or PPFT to update particle origins and orientations
    if($auto_param{have_map}) {
	--$auto_param{niter}; # Decrement number of iterations
	printf OUT "*** Updating origins and orientations\n\n";

	if($auto_param{mode} eq 'refine') {
	    $input_file  = "po2r_input_iter_${iter}";
	    $output_file = "po2r_output_iter_${iter}";
	    $po2r_param{in_3d_name} = $auto_param{start_map};
	    unless(-r $po2r_param{in_3d_name}) {
		printf OUT "PO2R unable to open $po2r_param{in_3d_name}\n";
		printf OUT "Exiting AUTO3DEM\n";
		exit;
	    }

	    if($auto_param{hollow_map}) {
		$hollow_param{in_3d_name} = $po2r_param{in_3d_name};
		$po2r_param{in_3d_name} = $hollow_param{out_file_name};
		make_program_input('hollow_in', 'pcut', \%hollow_param);
		run_mpi_prog($pcut_param{bin}, 'hollow_in',
				$pcut_cpu, 'hollow_out', $nodefile);
	    }

	    make_program_input($input_file, 'po2r', 
			       \%po2r_param, \@ptlefiles);
	    run_mpi_prog($po2r_param{bin},  $input_file,       
			 $po2r_cpu, $output_file, $nodefile);
	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	} else {
	    $ppft_param{model_filename} = $auto_param{start_map};
	    unless(-r $ppft_param{model_filename}) {
		printf OUT "PPFT unable to open $ppft_param{model_filename}\n";
		printf OUT "Exiting AUTO3DEM\n";
		exit;
	    }

	    if($auto_param{hollow_map}) {
		$hollow_param{in_3d_name} = $ppft_param{model_filename};
		$ppft_param{model_filename} = $hollow_param{out_file_name};
		make_program_input('hollow_in', 'pcut', \%hollow_param);
		run_mpi_prog($pcut_param{bin}, 'hollow_in',
				$pcut_cpu, 'hollow_out', $nodefile);
	    }

	    $input_file  = "ppft_input_iter_${iter}";
	    $output_file = "ppft_output_iter_${iter}";
	    $ppft_param{pftrads_filename} = "ppft_iter_${iter}.rads";
	    $ppft_param{pftres1_filename} = "ppft_iter_${iter}.res1";
	    $ppft_param{pftres2_filename} = "ppft_iter_${iter}.res2";
	    if(-e $ppft_param{pftrads_filename}) {
		unlink $ppft_param{pftrads_filename};
	    }
	    if(-e $ppft_param{pftres1_filename}) {
		unlink $ppft_param{pftres1_filename};
	    }
	    if(-e $ppft_param{pftres2_filename}) {
		unlink $ppft_param{pftres2_filename};
	    }
	    make_program_input($input_file, 'ppft', 
			       \%ppft_param, \@ptlefiles);
	    run_mpi_prog($ppft_param{bin},  $input_file,       
			 $ppft_cpu, $output_file, $nodefile);
	    
	    # Parse PPFT output to update inner and outer radii of the
	    # annulus. Use these to set (annulus_low, annulus high)
	    # for next run of PPFT and (in_rad, out_rad) for PCUT
	    unless($auto_param{freeze_annulus}) {
		my ($ann_lo, $ann_hi) = 
		    get_ann_lo_hi($ppft_param{pftrads_filename},
				  $ppft_param{pftrad_lo},
				  $ppft_param{pftrad_step},
				  $auto_param{boxrad},
				  $ppft_param{bin_factor});
		$ppft_param{annulus_low} = $ann_lo;
		$ppft_param{annulus_high} = $ann_hi;
		$pcut_param{in_rad} = $ann_lo;
		$pcut_param{out_rad} = $ann_hi;
		print OUT "   PPFT annulus_low changed to  $ann_lo\n";
		print OUT "   PPFT annulus_high changed to $ann_hi\n";
		print OUT "   PCUT in_rad changed to       $ann_lo\n";
		print OUT "   PCUT out_rad changed to      $ann_hi\n";
	    }

	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	    move ($ppft_param{pftrads_filename}, 
		  "$intfiles_dir/$ppft_param{pftrads_filename}");
	    move ($ppft_param{pftres1_filename}, 
		  "$intfiles_dir/$ppft_param{pftres1_filename}");
	    move ($ppft_param{pftres2_filename}, 
		  "$intfiles_dir/$ppft_param{pftres2_filename}");
	}

	# Increment the particle origin and orientation files
	incr_ptlefiles(\@ptlefiles);

	# Write out restart file
	my $restart_file = "$auto_param{outfile}_restart_${iter}a";
	$restart_file = "${startdir}/${restart_file}";
	$auto_param{have_map} = 0; # Upon restart, won't have map
	write_params($restart_file, \%auto_param, \%po2r_param,
		     \%p3dr_param, \%pcut_param, \%psf_param,
		     \%ppft_param, \%pctfr_param, \@ptlefiles);
    }

    print OUT "*** Evaluating selection criteria\n\n";
    
    my $res_lothresh_best = 1000;
    my $res_hithresh_best = 1000;
    my $reached_hithresh_best = 0;
    my $reached_lothresh_best = 0;
    my $selection_best = 0;
    my @spatial_freq;
    my @fsc;

    # Make a copy of P3DR hash for creating even and odd maps and set
    # the zero_fill to be one.
    my %p3dr_param_oddeven = %p3dr_param;
    $p3dr_param_oddeven{zero_fill} = 1.0;
    
    # LOOP OVER SELECTION CRITERIA

    # If auto_param{estimate_res} is false, then NO resolution
    # estimation will be performed and selection criterion 0 will be
    # used to construct the map

    my @jlist;
    if($auto_param{estimate_res}) {
	my $nso = $auto_param{nselect_offset};
	@jlist = (-$nso .. $nso);
    } else {
	$selection_best = 0;
	@jlist = ();
    }

    foreach my $j (@jlist) {
	
	my $suffix = "${iter}_${j}";
	$suffix =~ s/-/n/;
	
	# Generate the filtered particle files
	my ($nselect, @selectfiles) = 
	    gen_filtered_ptles(\@ptlefiles, \%auto_param, $j);
	
	# Run P3DR to generate even map
	$input_file  = "p3dr_even_input_iter_$suffix";
	$output_file = "p3dr_even_output_iter_$suffix";
	$p3dr_param_oddeven{oddeven} = 2;
	$p3dr_param_oddeven{out_file_name} = "map_even";
	make_program_input($input_file, 'p3dr', \%p3dr_param_oddeven, 
			   \@selectfiles);
	run_mpi_prog($p3dr_param_oddeven{bin}, $input_file, 
		     $p3dr_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	# Run P3DR to generate odd map
	$input_file  = "p3dr_odd_input_iter_$suffix";
	$output_file = "p3dr_odd_output_iter_$suffix";
	$p3dr_param_oddeven{oddeven} = 1;
	$p3dr_param_oddeven{out_file_name}  = "map_odd";
	make_program_input($input_file, 'p3dr', \%p3dr_param_oddeven, 
			   \@selectfiles);
	run_mpi_prog($p3dr_param_oddeven{bin}, $input_file,  
		     $p3dr_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	foreach my $selectfile (@selectfiles) {
	    move $selectfile, "$intfiles_dir/$selectfile";
	}

	# Run PCUT to cut the even map
	$input_file  = "pcut_even_input_iter_$suffix";
	$output_file = "pcut_even_output_iter_$suffix";
	$pcut_param{in_3d_name}    = "map_even";
	$pcut_param{out_file_name} = "map_even_cut";
	make_program_input($input_file, 'pcut', \%pcut_param);
	run_mpi_prog($pcut_param{bin}, $input_file, 
		     $pcut_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	# Run PCUT to cut the odd map
	$input_file  = "pcut_odd_input_iter_$suffix";
	$output_file = "pcut_odd_output_iter_$suffix";
	$pcut_param{in_3d_name}     = "map_odd";
	$pcut_param{out_file_name}  = "map_odd_cut";
	make_program_input($input_file, 'pcut', \%pcut_param);
	run_mpi_prog($pcut_param{bin}, $input_file,  
		     $pcut_cpu, $output_file, $nodefile);
	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	# Run PSF to generate FSC curve
	$input_file  = "psf_input_iter_$suffix";
	$output_file = "psf_output_iter_$suffix";
	$psf_param{fileeven} = "map_even_cut";
	$psf_param{fileodd}  = "map_odd_cut";
	make_program_input($input_file, 'psf', \%psf_param);
	run_mpi_prog($psf_param{bin},  $input_file,       
		     $psf_cpu, $output_file, $nodefile);

	# Parse the PSF output file to determine current resolution and
	# high resolution to be used in next iteration

	my ($res_hithresh, $res_lothresh, $reached_hithresh,
	    $reached_lothresh, $spatial_freq_ref, $fsc_ref) =
	    psf_parse($auto_param{fsc_hithresh},
		      $auto_param{fsc_lothresh}, $output_file);
	
	# Determine whether this selection criterion results in
	# the best resolution
	if($res_hithresh < $res_hithresh_best) {
	    $res_hithresh_best = $res_hithresh;
	    $res_lothresh_best = $res_lothresh;
	    $selection_best = $j;
	    $reached_hithresh_best = $reached_hithresh;
	    $reached_lothresh_best = $reached_lothresh;
	    @spatial_freq = @{$spatial_freq_ref};
	    @fsc = @{$fsc_ref};
	    copy($output_file, "$auto_param{outfile}_FSC_iter_${iter}");
	}

	move $input_file,  "$intfiles_dir/$input_file";
	move $output_file, "$intfiles_dir/$output_file";
	
	# Cleanup temporary map files
	unlink 'map_even', 'map_even_cut', 'map_odd', 'map_odd_cut';
    }

    my $fsctxt = "$auto_param{outfile}_FSC_iter_${iter}";
    my $fscpng = "${fsctxt}.png";
    
    # Construct FSC plot
    if($gdflag) {
	my @data = (\@spatial_freq, \@fsc);
	my $graph = GD::Graph::lines->new(400, 300);
	$graph->set( x_label           => 'Spatial Frequency',
		     y_label           => 'FSC',
		     title             => "$auto_param{outfile} iter $iter",
		     t_margin          => 10,
		     b_margin          => 10,
		     l_margin          => 10,
		     r_margin          => 20,
		     x_tick_number     => 'auto',
		     y_min_value       => 0,
		     y_max_value       => 1.2,
		     y_tick_number     => 12,
		     y_label_skip      => 2 ) 
	    or die $graph->error;
	my $gd = $graph->plot(\@data) or die $graph->error;
	open(IMG, ">$fscpng") or die $!;
	binmode IMG;
	print IMG $gd->png;
    }
    
    # Send notification that iteration has completed and attach FSC curve
    if(defined $auto_param{recipient}) {
	my $subj = "FSC curve - iteration $iter";
	my @attach = ("$fsctxt");
	if($gdflag) {
	    push(@attach, "$fscpng");
	}
	my $msgfile = "message.txt";
	my $currentdir = getcwd();
	open(MSG, ">$msgfile");
	print MSG "Completed $auto_param{outfile} iteration $iter\n";
	print MSG "FSC data attached\n";
	close(MSG);
	$msgfile = "$currentdir/$msgfile";
	sendmail($msgfile, $subj, $auto_param{recipient}, @attach);
	unlink $msgfile;
    }

    print OUT "*** Construct map using selection criterion ";
    print OUT "$selection_best\n";

    # Adjust resolutions based on the best $res_lothresh
    if($auto_param{estimate_res} && !$auto_param{freeze_res}) {
	update_res(\%po2r_param, \%p3dr_param, \%psf_param,
		   \%ppft_param, \%pctfr_param, \%auto_param,
		   $pixel_size, $res_lothresh_best,
		   $reached_lothresh_best); 
    }

    # Generate the filtered particle files using best selection criterion
    my ($nselect, @selectfiles) = 
	gen_filtered_ptles(\@ptlefiles, \%auto_param, $selection_best);

    
    # If we have new particles that we are refining relative to an
    # existing higher resolution map, we do not need to construct a
    # new map.
    unless($auto_param{new_ptles}) {

	# Write out FSC data file to be read by P3DR
	my $fsc_file_name;
	if($auto_param{noise_suppression}) {
	    my $fsc_count = @fsc;
	    $fsc_file_name = "fscfile_${iter}";
	    open(FSC, ">$fsc_file_name");
	    print FSC "$fsc_count\n";
	    for(my $k=0; $k<$fsc_count; $k++) {
		printf FSC "%7.4f\t%7.4f\n", $spatial_freq[$k], $fsc[$k];
	    }
	    $p3dr_param{fsc_file_name} = $fsc_file_name;
	}
    
	# Run P3DR to generate full map using filtered particle files
	$p3dr_param{out_file_name} = "$map_iter";
	$input_file  = "p3dr_input_iter_${iter}";
	$output_file = "p3dr_output_iter_${iter}";
	$p3dr_param{oddeven} = 0;
	make_program_input($input_file, 'p3dr', \%p3dr_param, \@selectfiles);
	run_mpi_prog($p3dr_param{bin}, $input_file,      
		     $p3dr_cpu, $output_file, $nodefile);
	delete $p3dr_param{fsc_file_name};

	move $input_file,     "$intfiles_dir/$input_file";
	move $output_file,    "$intfiles_dir/$output_file";
	if($auto_param{noise_suppression}) {
	    move $fsc_file_name,  "$intfiles_dir/$fsc_file_name";
	}
	foreach my $selectfile (@selectfiles) {
	    move $selectfile, "$intfiles_dir/$selectfile";
	}

	# Run PCTFR to refine CTF parameters if we are in refine mode
	# and the refine_ctf flag has been set.

	if($auto_param{refine_ctf} && $auto_param{mode} eq 'refine' &&
	   $pctfr_param{ctfmode} > 0) {
	    $input_file  = "pctfr_input_iter_${iter}";
	    $output_file = "pctfr_output_iter_${iter}";
	    $pctfr_param{in_3d_name}    = $map_iter;
	    make_program_input($input_file, 'pctfr', 
			       \%pctfr_param, \@ptlefiles);
	    run_mpi_prog($pctfr_param{bin}, $input_file,      
			 $pctfr_cpu, $output_file, $nodefile);
	    move $input_file,  "$intfiles_dir/$input_file";
	    move $output_file, "$intfiles_dir/$output_file";
	}

	# Set start_map to be used for next iteration or run
	$auto_param{start_map} = "$map_iter";
    }

    # Write out summary information
    my $delta_angle;
    if($auto_param{mode} eq "refine") {
	$delta_angle =  $po2r_param{dangle};
    } else {
	$delta_angle = $ppft_param{delta_theta};
    }
    my $flag = " ";
    unless($reached_hithresh_best) {
	$flag = "*";
    }

    my $time_iter_end= time();
    my $time_iter = $time_iter_end - $time_iter_start;

    my $outmode = $auto_param{mode};
    if($outmode eq 'search') {
	$outmode .= "($ppft_param{bin_factor})";
    }

    printf SUM ("%4d %9s %6.2f%1s %6.2f  %7.2f %9.2f %5d  %6d %4d\n",
		$iter, $outmode, $res_hithresh_best, $flag,
		$delta_angle, $p3dr_param{res_min}, $p3dr_param{res_max},
		$time_iter, $nselect, $ncpu);

    # Add map file to list
    push(@mapfiles, $map_iter);

    # Add estimated resolution to list
    unshift(@estres, $res_hithresh_best);

    # Test progress and switch from search to refine mode if
    #   (1) Current mode is search
    #   (2) Resolution improved by less than 0.25A
    #   (3) switch_mode flag is set
    #   (4) PPFT bin_factor == 1 or bin_reduce not set

    if(defined $estres[0] && defined $estres[1] &&
       $auto_param{mode} eq 'search' && 
       ($estres[1]-$estres[0]) < 0.25 &&
       $auto_param{switch_mode} &&
       ($ppft_param{bin_factor} == 1 || !$auto_param{bin_reduce})) {
	$auto_param{mode} = 'refine';
    }
    
    # Test progress and reduce bin factor if
    #   (1) Current mode is search
    #   (2) Resolution improved by less than 0.25A
    #   (3) PPFT bin_factor > 1
    #   (4) bin_reduce flag set

    if(defined $estres[0] && defined $estres[1] &&
       $auto_param{mode} eq 'search' && 
       ($estres[1]-$estres[0]) < 0.25 &&
       $ppft_param{bin_factor} > 1 && $auto_param{bin_reduce}) {
	--$ppft_param{bin_factor};
    }

    # Write out restart file
    my $restart_file = "$auto_param{outfile}_restart_${iter}b";
    $restart_file = "${startdir}/${restart_file}";
    $auto_param{have_map} = 1; # Upon restart, will have map
    write_params($restart_file, \%auto_param, \%po2r_param,
		 \%p3dr_param, \%pcut_param, \%psf_param,
		 \%ppft_param, \%pctfr_param, \@ptlefiles);


    # Quit early if flag set and FSC curve does not drop below 0.5
    if($res_hithresh_best <= $psf_param{res_max} &&
       $auto_param{quit_early}) {
	print "\n--- Terminating auto3dem early ---\n";
	print "Starting model successfully calculated\n\n"; 
	print "Note that this message should only be encountered when\n";
	print "using the random model method. If you see this message\n";
	print "during an ordinary auto3dem run, remove the line\n";
	print "auto \"quit_early 1\" from your auto3dem master file\n\n";
	print OUT "\n--- Terminating auto3dem early ---\n";
	print OUT "Starting model successfully calculated\n\n"; 
	print OUT "Note that this message should only be encountered when\n";
	print OUT "using the random model method. If you see this message\n";
	print OUT "during an ordinary auto3dem run, remove the line\n";
	print OUT "auto \"quit_early 1\" from your auto3dem master file\n\n";
	chdir($startdir); # Make sure file written in correct place
	open(DONE, ">RMC_DONE");
	print DONE "DONE - starting model successfully calculated\n";
	close(DONE);
	chdir($auto_param{rundir});
	last;
    }

}

# Delete intermediate map files if flag set
if($auto_param{delete_maps}) {
    pop @mapfiles; # Pop the last file added off the list
    foreach my $map (@mapfiles) {
	unlink $map;
    }
}

# Get rid of stdin_temp file (generated by PPFT)
unlink "stdin_temp";

# Return to starting directory
chdir($startdir);

# Move restart files into their own directory
my $restart_dir = "$auto_param{outfile}_RESTARTS";
mkdir $restart_dir;
foreach my $file (<$auto_param{outfile}_restart*>) {
    move $file, $restart_dir;
}

# Restore number of iterations to original value
$auto_param{niter} = $niter_original;

my $continue_file = "$auto_param{outfile}_continue";
write_params($continue_file, \%auto_param, \%po2r_param,
	     \%p3dr_param, \%pcut_param, \%psf_param,
	     \%ppft_param, \%pctfr_param, \@ptlefiles);

my $time_stop = time();
my $time_elapsed = $time_stop - $time_start;
print OUT "Auto3dem finished\n";
print OUT "   Total run time = $time_elapsed seconds\n";

close(OUT);
close(SUM);
